/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/


/*
 *  Here are some design insights on the wrapper layer:
 *
 *  One thread (demux) is actually receiving messages and notifications from the socket object.
 * When a message is received, along with the stream it is received from, it is saved 
 * in the buffer queue (FIFO dynamic lists of buffers) corresponding to this stream.
 * 
 *  The "pull" callback that the gnutls library expects for transport will actually pull from the
 * buffer queue. Conditional variables are used to block and wakeup when messages are received on
 * the appropriate stream.
 *  
 *  The "push" callback is more straight-forward.
 *
 *  Once the handshake is complete, one thread per stream is created and looping on the 
 * gnutls_record_recv function, using the appropriate gnutls session. When a message is
 * received successfuly, it is queued in the global FIFO for decrypted messages. The
 * tls_sctpw_recv function picks up messages from this queue.
 */


#include <pthread.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/sctp.h>
#include <sys/uio.h>

#ifdef STANDALONE_WRAPPER

# include "gnutls_sctp_wrapper.h"

/* define to anything to issue debug messages */
static int is_client = 0;

# define TRACE_DEBUG(level,format,args... ) \
	printf("%s( %-15s): " format "\n", is_client?"                   [CLI] ":"[SRV] ", __FUNCTION__, ## args);
/* define to anything to issue trace messages */
# define TRACE_ENTRY(_format,_args... )	\
	TRACE_DEBUG(1, "Enter (" _format ")" , ##_args );

#else /* STANDALONE_WRAPPER */

# include "sec_tls_gnutls.h"

#endif /* STANDALONE_WRAPPER */

/****************************************************************************
*****************************************************************************
**        Data structures (internal to this module)                        **
*****************************************************************************
****************************************************************************/

/* A buffer received on the socket, with its meta-data */
typedef struct  {
	unsigned char * data;	 /* Pointer to the allocated data */
	size_t 		len;	 /* Length of the data */
	size_t 		offset;	 /* start of unread data */
	int    		streamid;/* The stream id on which the buffer was received */
} buffer_t;

/* An item from a buffer queue list */
typedef struct _bqi {
	buffer_t	 buf;	/* The actual buffer data and meta-data */
	struct _bqi	*next;	/* Next buffer in the queue */
	struct _bqi	*prev;	/* Previous buffer in the queue */
} bqi_t;

/* A buffer queue head */
typedef struct {
	bqi_t		sentinel;	/* a dummy element to initialize the chain */
	pthread_mutex_t	mtx;		/* mutex to protect the list */
	pthread_cond_t	cond;		/* condvar to signal when new message is posted */
	int		count;		/* The number of messages in the queue */
} bq_t;

/* To hold session state for session resuming */
typedef struct {
	char 		key[GNUTLS_MAX_SESSION_ID];
	size_t 		keylen;
	gnutls_datum_t 	data;
} sr_t;

/* forward declaration */
struct _tls_sctpw_ctx;

/* For each pair of streams in the association, we need the following data: */
typedef struct {
	int		 	 stream; /* The stream id of this object. This is also the index in the array */
	struct _tls_sctpw_ctx	*ctx;	 /* Pointer to the head of the context object */
	bq_t		 	 queue;	 /* buffer queue of the encrypted chunks of data received and demux'd */
	gnutls_session_t 	 session;/* The gnutls session corresponding to this pair of streams. */
	sr_t			*resume; /* pointer to data to resume a session */
	pthread_t		 th;	 /* After handshake, the thread that receives and decrypts the data */
	int			 th_st;  /* Is the thread created? 0: no; 1: yes, running; 2: yes, terminated */
} pairinfo_t;

/* State of the wrapper context */
typedef enum {
	SCTPW_CTX_INIT,	/* The context has been initialized */
	SCTPW_CTX_OPEN,	/* Handshake has been done */
	SCTPW_CTX_BROKEN/* An error occurred on the socket */
} ctx_state;

/* Header of the context structure, that contain the data unique to each association */
typedef struct {
	int		sock;	/* The SCTP socket */
	int		nbstrm;	/* Number of streams to use */
	int		strauto;/* id of the stream to use when "-1" is specified */
	int		side;	/* 0: server, !0: client */
	ctx_state	state;	/* The state of this context */
	pthread_t	demux;	/* The demux thread */
	bq_t		rcvd;	/* queue of the deciphered chunks of data */
} ctx_hdr_t;

/* Now the definition of the wrapper context, corresponding to opaque type tls_sctpw_ctx: */
typedef struct _tls_sctpw_ctx {
	ctx_hdr_t	hdr;	 /* Header of the structure */
	pairinfo_t	pairs[]; /* array of hdr.nbstrm * pairinfo_t structures */
} tls_sctpw_ctx_t;


/****************************************************************************
*****************************************************************************
**       TLS over SCTP wrapper library -- to use GNU TLS over SCTP         **
*****************************************************************************
****************************************************************************/ 

/***** Macro to use when dumping sockaddr ******/
/* char addrbuf[INET6_ADDRLEN] must exist in the context 
 *  sa is an address of a sockaddr
 *  res is a char * that points to the string in the end.
 */
#define DUMP_SA( sa, res ) {						\
	struct sockaddr * _sa_ = (struct sockaddr *)(sa);		\
	struct sockaddr_in * _sin_ = (struct sockaddr_in *) (sa);	\
	struct sockaddr_in6 * _sin6_ = (struct sockaddr_in6 *) (sa);	\
	switch (_sa_->sa_family) {					\
		case AF_INET:						\
			res = (char *)inet_ntop(AF_INET, 		\
					&(_sin_->sin_addr.s_addr), 	\
					addrbuf, 			\
					sizeof(addrbuf));		\
			break;						\
		case AF_INET6:						\
			res = (char *)inet_ntop(AF_INET6, 		\
					&(_sin6_->sin6_addr.s6_addr), 	\
					addrbuf, 			\
					sizeof(addrbuf));		\
			break;						\
		default:						\
			snprintf(addrbuf, 				\
				 sizeof(addrbuf), 			\
				 "Unknown AF: %d", 			\
				 _sa_->sa_family);			\
			res = (char *)addrbuf;				\
	}								\
}


/***** Buffer Queues Management ******/

#define BQI_LINK_INIT( _item_ ) { 				\
	((bqi_t *)( _item_ ))->next = (bqi_t *)( _item_ );	\
	((bqi_t *)( _item_ ))->prev = (bqi_t *)( _item_ );	\
}

#define BQ_PUSH( bq, bqi ) {				\
	bqi_t * _sent_ = &(((bq_t *)( bq ))->sentinel);	\
	bqi_t * _new_  = (bqi_t *)( bqi );		\
	_new_->next = _sent_;				\
	_new_->prev = _sent_->prev;			\
	_sent_->prev->next = _new_;			\
	_sent_->prev=_new_;				\
	((bq_t *)( bq ))->count++;			\
}
#define BQ_POP( bq, bqi ) {				\
	bqi_t * _rem_  = (bqi_t *)( bqi );		\
	_rem_->next->prev = _rem_->prev;		\
	_rem_->prev->next = _rem_->next;		\
	BQI_LINK_INIT(_rem_);				\
	((bq_t *)( bq ))->count--;			\
}

/* Initialize a queue */
static int bq_init(bq_t * bq)
{
	int ret = 0;
	
	memset(bq, 0, sizeof(bq_t));
	BQI_LINK_INIT( &bq->sentinel );
	
	ret = pthread_mutex_init(&bq->mtx, NULL);
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_init failed: %s", strerror(ret));
		return ret;
	}
	
	ret = pthread_cond_init(&bq->cond, NULL);
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_cond_init failed: %s", strerror(ret));
		return ret;
	}
	
	return 0;
}

/* in case of cancellation */
static void bq_cleanup(void * _bq)
{
	int ret;
	bq_t * bq = (bq_t *)_bq;
	
	TRACE_ENTRY( "%p", bq );
	
	/* Now unlock the queue, and we're done */
	ret = pthread_mutex_unlock( &bq->mtx );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_unlock failed: %s", strerror(ret));
	}
	
	/* End of cleanup handler */
	return;
}

/* Post a message in a queue (we use directly a bqi_t to avoid re-allocation) */
static int bq_post(bq_t * bq, bqi_t *buf)
{
	int ret = 0;
	
	if ((bq == NULL) || (buf == NULL)) {
		TRACE_DEBUG(INFO, "Invalid parameters");
		return EINVAL;
	}
	
	/* lock the queue */
	ret = pthread_mutex_lock(&(bq->mtx));
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_lock failed: %s", strerror(ret));
		return ret;
	}
	
	/* Push the element */
	BQ_PUSH(bq, buf);
	
	/* Signal the condition */
	ret = pthread_cond_signal(&(bq->cond));
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_cond_signal failed: %s", strerror(ret));
		return ret;
	}
	
	/* unlock the queue */
	ret = pthread_mutex_unlock(&(bq->mtx));
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_unlock failed: %s", strerror(ret));
		return ret;
	}
	
	return 0;
}

/* Wait and retrieve a complete message from a queue */
static int bq_get(bq_t * bq, bqi_t **buf)
{
	int ret = 0;
	
	if ((bq == NULL) || (buf == NULL)) {
		TRACE_DEBUG(INFO, "Invalid parameters");
		return EINVAL;
	}
	
	/* lock the queue */
	ret = pthread_mutex_lock(&(bq->mtx));
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_lock failed: %s", strerror(ret));
		return ret;
	}
	
recheck:
	if (bq->count > 0) {
		*buf = bq->sentinel.next;
		BQ_POP( bq, *buf );
	} else {
		/* wait for a message to be posted */
		pthread_cleanup_push(bq_cleanup, bq);
		ret = pthread_cond_wait(&(bq->cond), &(bq->mtx));
		pthread_cleanup_pop(0);
		if (ret != 0) {
			TRACE_DEBUG(INFO, "pthread_cond_wait failed: %s", strerror(ret));
			return ret;
		}
		/* Need to recheck in case of spurious wake up */
		goto recheck;
	}
	
	/* unlock the queue */
	ret = pthread_mutex_unlock(&(bq->mtx));
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_unlock failed: %s", strerror(ret));
		return ret;
	}
	
	return 0;
}

/* Read data into a provided buffer (len is in/out parameter) */
static int bq_read(bq_t * bq, void * data, size_t *len)
{
	int ret = 0;
	bqi_t *bqi = NULL;
	
	if ((bq == NULL) || (data == NULL) || (len == NULL)) {
		TRACE_DEBUG(INFO, "Invalid parameters");
		return EINVAL;
	}
	
	/* lock the queue */
	ret = pthread_mutex_lock(&(bq->mtx));
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_lock failed: %s", strerror(ret));
		return ret;
	}
	
recheck:
	if (bq->count > 0) {
		/* Now we have data to read */
		bqi = bq->sentinel.next;
		if (*len >= bqi->buf.len - bqi->buf.offset) {
			/* we can read the whole item at once */
			BQ_POP( bq, bqi );
			memcpy(data, bqi->buf.data + bqi->buf.offset, bqi->buf.len - bqi->buf.offset);
			*len = bqi->buf.len - bqi->buf.offset;
			/* free the bqi, we don't need it anymore */
			free(bqi->buf.data);
			free(bqi);
		} else {
			/* supplied buffer is too small, read only partial buffer */
			memcpy(data, bqi->buf.data + bqi->buf.offset, *len);
			bqi->buf.offset += *len;
		}
	} else {
		/* wait for a message to be posted */
		pthread_cleanup_push(bq_cleanup, bq);
		ret = pthread_cond_wait(&(bq->cond), &(bq->mtx));
		pthread_cleanup_pop(0);
		if (ret != 0) {
			TRACE_DEBUG(INFO, "pthread_cond_wait failed: %s", strerror(ret));
			return ret;
		}
		/* Need to recheck in case of spurious wake up */
		goto recheck;
	}
	
	/* unlock the queue */
	ret = pthread_mutex_unlock(&(bq->mtx));
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_unlock failed: %s", strerror(ret));
		return ret;
	}
	
	return 0;
}

/* Destroy a queue (must be empty) */
static int bq_fini(bq_t * bq)
{
	int ret;
	
	if (bq == NULL) {
		TRACE_DEBUG(INFO, "Invalid parameters");
		return EINVAL;
	}
	
	while (bq->count) {
		bqi_t * del = NULL;
		ret = bq_get(bq, &del);
		if (ret != 0) {
			TRACE_DEBUG(INFO, "bq_get failed: %s", strerror(ret));
			return ret;
		}
		free(del->buf.data);
		free(del);
	}
	
	ret = pthread_cond_destroy(&(bq->cond));
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_cond_destroy failed: %s", strerror(ret));
		return ret;
	}
	
	ret = pthread_mutex_destroy(&(bq->mtx));
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_destroy failed: %s", strerror(ret));
		return ret;
	}
	
	return 0;
}

/***** Push and pull functions ******/

/* Send data over the connection, called by gnutls */
static ssize_t sctpw_push (gnutls_transport_ptr_t pair, const void * data, size_t len)
{
	int sock;
	int stream;
	
	/* TRACE_ENTRY("%p %p %d", pair, data, len); */
	
	if ((pair == NULL) || (data == NULL)) {
		TRACE_DEBUG(INFO, "Received invalid NULL parameter");
		return -1;
	}
	
	sock   = ((pairinfo_t *)pair)->ctx->hdr.sock;
	stream = ((pairinfo_t *)pair)->stream;
		
	return tls_sctp_send(sock, stream, data, len);
}

/* Retrieve data received on a stream and demultiplexed already */
static ssize_t sctpw_pull (gnutls_transport_ptr_t pair, void * buf, size_t buflen)
{
	int ret = 0;
	size_t len = buflen;
	
	/* TRACE_ENTRY("%p %p %d", pair, buf, buflen); */
	
	ret = bq_read( &(((pairinfo_t *)pair)->queue), buf, &len );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "bq_read failed: %s", strerror(ret));
		return -1;
	}
	
	return len;
}

/* Read data on the socket and queue it to the appropriate buffer queue */
static void * sctpw_demux (void * arg)
{
	tls_sctpw_ctx_t * ctx = (tls_sctpw_ctx_t *) arg;
	ssize_t ret;
	int 	reti;
	tls_sctp_recvevents event;
	bqi_t * new;
	
	/* loop forever; the thread will be canceled when we want to terminate */
	while (1) {
		/* Prepare a structure to save the buffer */
		new = (bqi_t *) malloc(sizeof(bqi_t));
		if (new == NULL) {
			TRACE_DEBUG(INFO, "malloc failed");
			return NULL;
		}
		memset(new, 0, sizeof(bqi_t));
		BQI_LINK_INIT(new);
		
		/* Now receive data from the SCTP socket */
		ret = tls_sctp_recv( ctx->hdr.sock, &event, &(new->buf.streamid), (void *)&(new->buf.data), &(new->buf.len), 4096 );
		if (ret <= 0)
			free(new);
			
		if (ret < 0) {
			/* The socket is no more valid */
			TRACE_DEBUG(INFO, "tls_sctp_recv failed");
			goto error;
		}
		
		if ((ret == 0) && ((event == RCV_ERROR) || (event == RCV_SHUTDOWN))) {
			/* The socket is no more valid */
			TRACE_DEBUG(INFO, "tls_sctp_recv received a termination notification");
			goto error;
		}
		
		if (ret == 0)
			continue;
		
		if (new->buf.streamid >= ctx->hdr.nbstrm) {
			TRACE_DEBUG(INFO, "Received data on a non-protected stream, discarding.");
			free(new);
			continue;
		}
			
		/* ok we have received a real message, demux and save it for sctpw_pull */
		reti = bq_post( &(ctx->pairs[new->buf.streamid].queue), new);
		if (reti != 0) {
			TRACE_DEBUG(INFO, "bq_post failed");
			goto error;
		}
	}
error:
				
	/* we should signal a control thread here */
	
	ctx->hdr.state = SCTPW_CTX_BROKEN;

	return NULL;

}

/* For session resuming in the server */
static int sr_store (void *dbf, gnutls_datum_t key, gnutls_datum_t data)
{
	sr_t *sr = (sr_t *)dbf;
	int ret = 0;
	
	/* We don't really do anything here, the session we are interested in is already stored. */
	/* we just return 0 if it is the same data, and -1 otherwise */
	
	if ((key.size != sr->keylen) ||	memcmp (key.data, sr->key, key.size))
		ret = -1;
		
	if ((data.size != sr->data.size) || memcmp (data.data, sr->data.data, data.size))
		ret = -1;
	
	return ret;
}
static int sr_remove (void *dbf, gnutls_datum_t key)
{
	/* sr_t *sr = (sr_t *)dbf; */
	
	/* Again, we don't do anything */
	return -1;
}
static gnutls_datum_t sr_fetch (void *dbf, gnutls_datum_t key)
{
	gnutls_datum_t res = { NULL, 0 };
	sr_t *sr = (sr_t *)dbf;
	
	/* This is the only db function we use, to return the session of pair[0] if the id matches */
	
	/* Check if the session matches the session of pair 0; we don't resume other sessions */
	if ((key.size == sr->keylen) && memcmp (key.data, sr->key, key.size) == 0)
	{
		/* It's the same id, so we can actually resume */
		res.size = sr->data.size;

		res.data = gnutls_malloc (res.size);
		if (res.data == NULL)
			return res;

		memcpy (res.data, sr->data.data, res.size);
	}
	return res;
}

/* Perform session resuming */
static void * sctpw_sess_resume (void * arg)
{
	int ret = 0;
	pairinfo_t * pair = (pairinfo_t *)arg;
	ctx_hdr_t * hdr;

	
	hdr = &(pair->ctx->hdr);
	
	if (hdr->side) {
		/* We are client side */
		gnutls_transport_ptr_t ptr_bkp;
		
		/* Backup the transport pointer of the session */
		ptr_bkp = gnutls_transport_get_ptr(pair->session);

		/* Use the data from the session on stream 0 to perform resuming */
		ret = gnutls_session_set_data(pair->session, pair->resume->data.data, pair->resume->data.size);
		if (ret != GNUTLS_E_SUCCESS) {
			TRACE_DEBUG(INFO, "gnutls_session_set_data failed on stream %d: %s", pair->stream, gnutls_strerror(ret));
			goto error;
		} else {
			TRACE_DEBUG(FULL, "data set: %llx...(%d bytes)", *(long long *)pair->resume->data.data, pair->resume->data.size);
		}
		
		/* Now restore the transport pointer */
		gnutls_transport_set_ptr(pair->session, ptr_bkp);

	} else {
		/* We are server side */
		
		/* The resume will be trigged during handshake. We provide the "fetch" function for this */
		gnutls_db_set_retrieve_function(pair->session, sr_fetch);
		
		/* The following functions are not useful but gnutls seems to need them... */
		gnutls_db_set_remove_function (pair->session, sr_remove);
		gnutls_db_set_store_function (pair->session, sr_store);
		
		/* The functions need this argument: */
		gnutls_db_set_ptr (pair->session, (void *)pair->resume);
	}
	
	/* And finaly perform the handshake */
	TRACE_DEBUG(FULL, "Starting handshake on stream %d", pair->stream);
	ret = gnutls_handshake( pair->session );
	if (ret != GNUTLS_E_SUCCESS) {
		TRACE_DEBUG(INFO, "Resumed handshake failed on stream %d: %s", pair->stream, gnutls_strerror(ret));
		goto error;
	}
	
	TRACE_DEBUG(FULL, "Handshake complete on stream %d, %s.", pair->stream, gnutls_session_is_resumed(pair->session) ? "session resumed" : "session NOT resumed");
	
	return (void *)0;
error:
	return (void *)-1;
}
	

/* Receive data from a pair of streams, once TLS handshake is done.  One thread per pair of streams */
#define BUF_SIZE	2048
static void * sctpw_recv_data (void * arg)
{
	int ret = 0;
	pairinfo_t * pair = (pairinfo_t *)arg;
	ctx_hdr_t * hdr;
	bqi_t * new;
	
	hdr = &(pair->ctx->hdr);
	
	while (1) {
		new = (bqi_t *) malloc(sizeof(bqi_t));
		if (new == NULL) {
			TRACE_DEBUG(INFO, "Memory allocation failed");
			goto term;
		}
		memset(new, 0, sizeof(bqi_t));
		BQI_LINK_INIT( new );
		new->buf.streamid = pair->stream;
		
		new->buf.data = malloc( BUF_SIZE );
		if (new->buf.data == NULL) {
			TRACE_DEBUG(INFO, "Memory allocation failed");
			goto term;
		}
		
		memset(new->buf.data, 0, BUF_SIZE );
		
		ret = gnutls_record_recv ( pair->session, new->buf.data, BUF_SIZE );
		if (ret <= 0) {
			TRACE_DEBUG(INFO, "gnutls_record_recv (stream %d) failed (%d)", pair->stream, ret);
			goto term;
		}
		
		new->buf.len = ret;
		
		/* Save the decrypted data in hdr->rcvd */
		ret = bq_post( &(hdr->rcvd), new);
		if (ret != 0) {
			TRACE_DEBUG(INFO, "bq_post failed: %s", strerror(ret));
			goto term;
		}
		
	}
term:
	if (ret != 0) {
		/* We should signal a control thread here... */
		TRACE_DEBUG(FULL, "Receiving thread of pair %d terminating...", pair->stream);
	}
	pair->th_st = 2;
	return NULL;
}

/* terminating / destroying a session */		
static int _tls_sctpw_end( tls_sctpw_ctx ** sctx, int graceful )
{
	int ret = 0;
	int i;
	tls_sctpw_ctx_t * _sctx;
	
	TRACE_ENTRY("%p", sctx);
	
	/* check parameter. for close call, only OPEN objects are valid */
	if ((sctx == NULL) || (*sctx == NULL)) {
		TRACE_DEBUG(INFO, "Invalid parameter");
		return EINVAL;
	}
	_sctx = (tls_sctpw_ctx_t *) *sctx;
	
	if (graceful && (_sctx->hdr.state != SCTPW_CTX_OPEN)) {
		/* force destroy here */
		graceful = 0;
	}
	
	/* First, stop the demux thread */
	if (_sctx->hdr.state != SCTPW_CTX_BROKEN) {
		ret = pthread_cancel(_sctx->hdr.demux);
		if (ret != 0) {
			TRACE_DEBUG(INFO, "pthread_cancel failed: %s", strerror(ret));
			return ret;
		}
	}

	ret = pthread_join(_sctx->hdr.demux, NULL);
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_join failed: %s", strerror(ret));
		return ret;
	}
	
	/* for each pair */
	for ( i= 0; i < _sctx->hdr.nbstrm; i++) {
		/* Shutdown the GNU TLS session */
		if (graceful) {
			ret = gnutls_bye(_sctx->pairs[i].session, GNUTLS_SHUT_WR);
			if (ret != GNUTLS_E_SUCCESS) {
				TRACE_DEBUG(INFO, "gnutls_bye returned: %s", gnutls_strerror(ret));
				/* we continue anyway */
			}
		}
		
		/* Terminate (cancel) the thread */
		if (_sctx->pairs[i].th_st) {
			
			if (_sctx->pairs[i].th_st == 1) {
				ret = pthread_cancel(_sctx->pairs[i].th);
				if (ret != 0) {
					TRACE_DEBUG(INFO, "pthread_cancel failed: %s", strerror(ret));
					return ret;
				}
			}
			
			ret = pthread_join(_sctx->pairs[i].th, NULL);
			if (ret != 0) {
				TRACE_DEBUG(INFO, "pthread_join failed: %s", strerror(ret));
				return ret;
			}
			
			_sctx->pairs[i].th_st = 0;
			
			/* Destroy the gnutls session */
			gnutls_deinit (_sctx->pairs[i].session);
		}
		
		/* Empty and destroy the queue */
		ret = bq_fini( &(_sctx->pairs[i].queue) );
		if (ret != 0) {
			TRACE_DEBUG(INFO, "bq_fini failed: %s", strerror(ret));
			return ret;
		}
	}
		
	/* Destroy the global objects */
	ret = bq_fini(&(_sctx->hdr.rcvd) );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "bq_fini failed: %s", strerror(ret));
		return ret;
	}
			
	free(_sctx);
	*sctx = NULL;
	
	return 0;
}

/**********************
 * External functions 
 *********************/

/* Initialize a wrapper context */
int tls_sctpw_init( int sock, int streams, tls_sctpw_ctx ** sctx, int client, gnutls_credentials_type_t credtype, void ** creds, int (*priocb)(gnutls_session_t *) )
{
	int ret = 0, i;
	tls_sctpw_ctx_t * _sctx = NULL;
	
	TRACE_ENTRY("%d %d %p %d", sock, streams, sctx, client);
	
	/* Check parameters */
	if ((sock < 0) || (streams < 0) || (sctx == NULL)) {
		TRACE_DEBUG(INFO, "Invalid parameters");
		return EINVAL;
	}
	
	/* Allocate storage for the new context object */
	_sctx = (tls_sctpw_ctx_t *) malloc( sizeof(ctx_hdr_t) + streams * sizeof(pairinfo_t) );
	if (_sctx == NULL) {
		ret = errno;
		TRACE_DEBUG(INFO, "Memory allocation failed: %s", strerror(ret));
		return ret;
	}
	
	memset(_sctx, 0, sizeof(ctx_hdr_t) + streams * sizeof(pairinfo_t));
	
	/* Initialize the header */
	_sctx->hdr.sock   = sock;
	_sctx->hdr.nbstrm = streams;
	_sctx->hdr.side   = client;
	_sctx->hdr.state  = SCTPW_CTX_INIT;
	/* _sctx->hdr.demux The demux thread is created in the end of this function */
	ret = bq_init( &(_sctx->hdr.rcvd) );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "bq_init failed: %s", strerror(ret));
		free(_sctx);
		return ret;
	}
	
	/* Now initialize all pairs */
	for (i = 0; i < streams; i++) {
		_sctx->pairs[i].stream = i;
		_sctx->pairs[i].ctx    = _sctx;
		
		ret = bq_init( &(_sctx->pairs[i].queue) );
		if (ret != 0) {
			TRACE_DEBUG(INFO, "bq_init failed: %s", strerror(ret));
			free(_sctx);
			return ret;
		}
		
		/* Now initialize the GNU TLS session for this pair of streams */
		ret = gnutls_init (&(_sctx->pairs[i].session), client ? GNUTLS_CLIENT : GNUTLS_SERVER);
		if (ret != GNUTLS_E_SUCCESS) {
			TRACE_DEBUG(INFO, "gnutls_init failed: %d", ret);
			return ret;
		}
		
		/* The priority setting of the session */
		if (priocb != NULL) {
			ret = (*priocb)(&(_sctx->pairs[i].session));
			if (ret != 0) {
				TRACE_DEBUG(INFO, "Setting priority with callback failed: %d", ret);
				return ret;
			}
		} else {
			ret = gnutls_set_default_priority (_sctx->pairs[i].session);
			if (ret != GNUTLS_E_SUCCESS) {
				TRACE_DEBUG(INFO, "gnutls_set_default_priority failed: %d", ret);
				return ret;
			}
		}
		
		/* The credentials settings of the session */
		ret = gnutls_credentials_set (_sctx->pairs[i].session, credtype, *creds);
		if (ret != GNUTLS_E_SUCCESS) {
			TRACE_DEBUG(INFO, "gnutls_credentials_set failed: %d", ret);
			return ret;
		}
		
		/* The transport settings of the session */
		gnutls_transport_set_ptr  (_sctx->pairs[i].session, (gnutls_transport_ptr_t) &(_sctx->pairs[i]));
		gnutls_transport_set_lowat(_sctx->pairs[i].session, 0 );
		gnutls_transport_set_pull_function(_sctx->pairs[i].session, sctpw_pull);
		gnutls_transport_set_push_function(_sctx->pairs[i].session, sctpw_push);
	
		/* The thread will be created at handshake time */
	}
	
	/* Now create the demux thread */
	ret = pthread_create( &(_sctx->hdr.demux), NULL, sctpw_demux, (void *)_sctx );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "Unable to create a thread: %s", strerror(ret));
		return ret;
	}
	
	*sctx = (tls_sctpw_ctx *) _sctx;
	
	return 0;
}

/* Handshake on all streams */
int tls_sctpw_handshake( tls_sctpw_ctx * sctx, int (*checkcb)(gnutls_session_t session, void * parm), void * parm )
{
	int ret = 0;
	int i, errs=0;
	sr_t resume_data;
	
	tls_sctpw_ctx_t * _sctx = (tls_sctpw_ctx_t *) sctx;
	
	TRACE_ENTRY("%p %p %p", sctx, checkcb, parm);
	
	if ((sctx == NULL) || (_sctx->hdr.state != SCTPW_CTX_INIT)) {
		TRACE_DEBUG(INFO, "Invalid parameter");
		return EINVAL;
	}
	
	/* First, perform a full handshake on the first pair of streams */
	ret = gnutls_handshake( _sctx->pairs[0].session );
	if (ret != GNUTLS_E_SUCCESS) {
		TRACE_DEBUG(INFO, "Full handshake failed on stream 0: %s", gnutls_strerror(ret));
		return ret;
	}

	TRACE_DEBUG(FULL, "Handskake complete on stream pair 0");

	if (checkcb != NULL) {
		/* Now verifying the credentials */
		ret = (*checkcb)(_sctx->pairs[0].session, parm);
		if (ret != 0) {
			TRACE_DEBUG(INFO, "The callback to verify the credentials returned an error: %d", ret);
			return ret;
		}
	}
	
	/* Ok, stream 0 is ready for use. Now retrieve the session data to perform session resuming */
	ret = gnutls_session_get_data2(_sctx->pairs[0].session, &resume_data.data);
	if (ret != GNUTLS_E_SUCCESS) {
		TRACE_DEBUG(INFO, "gnutls_session_get_data2 failed: %s", gnutls_strerror(ret));
		return ret;
	}
	/* We also need the session id (for the server) */
	resume_data.keylen = GNUTLS_MAX_SESSION_ID;
	ret = gnutls_session_get_id(_sctx->pairs[0].session, resume_data.key, &resume_data.keylen);
	if (ret != GNUTLS_E_SUCCESS) {
		TRACE_DEBUG(INFO, "gnutls_session_get_id failed: %s", gnutls_strerror(ret));
		return ret;
	}
	
	/* Resume the session on all other stream pairs, in parallel */
	for ( i= 1; i < _sctx->hdr.nbstrm; i++) {
		_sctx->pairs[i].resume = &resume_data;
		
		/* create the thread that resumes the session */
		ret = pthread_create( &(_sctx->pairs[i].th), NULL, sctpw_sess_resume, (void *)&(_sctx->pairs[i]));
		if (ret != 0) {
			TRACE_DEBUG(INFO, "pthread_create failed: %s", strerror(ret));
			return ret;
		}
		_sctx->pairs[i].th_st = 1;
	}
	
	/* Wait for the process to complete */
	for ( i= 1; i < _sctx->hdr.nbstrm; i++) {
		void * th_ret;
		
		/* create the thread that resumes the session */
		ret = pthread_join(_sctx->pairs[i].th, &th_ret);
		if (ret != 0) {
			TRACE_DEBUG(INFO, "pthread_join failed: %s", strerror(ret));
			return ret;
		}
		_sctx->pairs[i].th_st = 0;
		
		/* Check return value of the exchange */
		if ( th_ret != (void *)0 )
			errs ++ ;
	}
	
	/* Deallocate the resources of session resuming */
	gnutls_free(resume_data.data.data);
	
	/* If we had some errors: return an error */
	if (errs != 0) {
		TRACE_DEBUG(INFO, "Session could not be resumed on %d pairs of streams", errs);
		return ECONNABORTED;
	}
	
	/* For each pair of stream, now create the receiver thread */
	for ( i= 0; i < _sctx->hdr.nbstrm; i++) {
	
		/* create the thread that receives the data */
		ret = pthread_create( &(_sctx->pairs[i].th), NULL, sctpw_recv_data, (void *)&(_sctx->pairs[i]));
		if (ret != 0) {
			TRACE_DEBUG(INFO, "pthread_create failed: %s", strerror(ret));
			return ret;
		}
		_sctx->pairs[i].th_st = 1;
	}
	
	/* Mark the state as open now */
	_sctx->hdr.state  = SCTPW_CTX_OPEN;
		
	return 0;
}

/* Send data protected with TLS */
int tls_sctpw_send( tls_sctpw_ctx * sctx, int stream, unsigned char * data, size_t len )
{
	ssize_t sent;
	size_t rem = len;
	tls_sctpw_ctx_t * _sctx = (tls_sctpw_ctx_t *) sctx;
	
	TRACE_ENTRY("%p %d %p %d", sctx, stream, data, len);
	
	/* check parameters */
	if ((sctx == NULL) || (_sctx->hdr.state != SCTPW_CTX_OPEN)
	    || (stream >= _sctx->hdr.nbstrm)) {
		TRACE_DEBUG(INFO, "Invalid parameter");
		return EINVAL;
	}
	
	/* For automatic selection of a stream */
	if (stream < 0) {
		stream = _sctx->hdr.strauto;
	}
	/* This is not protected against multithreading, but it is not very important... */
	_sctx->hdr.strauto = ( stream + 1 ) % _sctx->hdr.nbstrm;
	
	/* now send the data */
	while (rem) {
		sent = gnutls_record_send(_sctx->pairs[stream].session, (const void *) data + (len - rem), rem);
		if (sent < 0) {
			TRACE_DEBUG(INFO, "gnutls_record_send returned %d: %s", sent, gnutls_strerror(sent));
			return sent;
		}
		rem -= sent;
	}
	
	return 0;
}

/* Receive protected data */
int tls_sctpw_recv( tls_sctpw_ctx * sctx, int *stream, void ** data, size_t *len )
{
	int ret = 0;
	
	tls_sctpw_ctx_t * _sctx = (tls_sctpw_ctx_t *) sctx;
	bqi_t * got = NULL;
	
	TRACE_ENTRY("%p %p %p %p", sctx, stream, data, len);
	
	/* check parameters */
	if ((data == NULL) || (len == NULL) || (stream == NULL) ||
	    (sctx == NULL) || (_sctx->hdr.state != SCTPW_CTX_OPEN)) {
		TRACE_DEBUG(INFO, "Invalid parameter");
		return EINVAL;
	}
	
	ret = bq_get( &(_sctx->hdr.rcvd), &got );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "bq_get failed: %s", strerror(ret));
		return ret;
	}
	
	*stream = got->buf.streamid;
	*len    = got->buf.len;
	*data   = got->buf.data;
	free(got);
	
	return 0;
}

int tls_sctpw_close( tls_sctpw_ctx ** sctx )
{
	TRACE_ENTRY("%p", sctx);
	return _tls_sctpw_end(sctx, 1);
}
	
int tls_sctpw_destroy( tls_sctpw_ctx ** sctx )
{
	TRACE_ENTRY("%p", sctx);
	return _tls_sctpw_end(sctx, 0);
}














/****************************************************************************
*****************************************************************************
**                TLS over SCTP wrapper test program                       **
*****************************************************************************
****************************************************************************/ 
#ifdef STANDALONE_WRAPPER

#include <signal.h>

#define TEST_PORT	4433
#define TEST_SRV_STR	4
#define TEST_CLI_STR	6

#define MSG0	"cli -> srv, stream 0"
#define MSG1	"srv -> cli, stream 2"
#define MSG2	"cli -> srv, stream -1"
#define MSG3	"srv -> cli, stream 3"

static int set_dh_prio(gnutls_session_t * session)
{
	const int kx_prio[] = { GNUTLS_KX_ANON_DH, 0 };
	
	gnutls_set_default_priority (*session);
	gnutls_kx_set_priority (*session, kx_prio);
	
	if (!is_client)
		gnutls_dh_set_prime_bits (*session, 1024);

	return 0;
}

/* The client function */
int client_process()
{
	int ret;
	struct sockaddr_in sin;
	int streams = TEST_CLI_STR;
	int sock;
	tls_sctpw_ctx *tls_ctx = NULL;
	gnutls_anon_client_credentials_t anoncred;

	TRACE_ENTRY();
	
	ret = gnutls_anon_allocate_client_credentials (&anoncred);
	if (ret != GNUTLS_E_SUCCESS) {
		TRACE_DEBUG(INFO,"Unable to initialize client credentials: %s", gnutls_strerror(ret));
		return -1;
	}
	
	/* Prepare the server information */
	memset(&sin, 0, sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_port   = htons(TEST_PORT);
	inet_pton(AF_INET, "127.0.0.1", &sin.sin_addr.s_addr);
	
	/* Now connect */
	TRACE_DEBUG(FULL, "Connecting to server ... ");
	ret = tls_sctp_connect( &sock, (struct sockaddr *)&sin, sizeof(sin), &streams );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "Failed to connect to server: %s", strerror(ret));
		return -1;
	}
	
	/* SCTP socket is ready */
	TRACE_DEBUG(FULL, "Connected; TLS can be run over %d streams", streams);
	
	/* test exchanging data without TLS */
	TRACE_DEBUG(FULL, "Sending data over SCTP streams 0 and 1 to test basic communication...");
	ret = tls_sctp_send(sock, 0, MSG0, strlen(MSG0));
	if (ret != strlen(MSG0)) {
		TRACE_DEBUG(INFO, "Partial send of data: %d / %d", ret, strlen(MSG0));
		return -1;
	}
	
	ret = tls_sctp_send(sock, 1, MSG2, strlen(MSG2));
	if (ret != strlen(MSG2)) {
		TRACE_DEBUG(INFO, "Partial send of data: %d / %d", ret, strlen(MSG0));
		return -1;
	}
	
	/* Initialize the wrapper context */
	TRACE_DEBUG(FULL, "Data sent; now creating wrapper context ...");
	ret = tls_sctpw_init( sock, streams, &tls_ctx, 1, GNUTLS_CRD_ANON, (void **)&anoncred, set_dh_prio );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "Unable to create wrapper context: %s", strerror(ret));
		return -1;
	}
	
	/* TLS handshake */
	TRACE_DEBUG(FULL, "Wrapper created, now initiating handshake ...");
	ret = tls_sctpw_handshake( tls_ctx, NULL, NULL );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "Handshake failed!");
		return -1;
	}
	
	/* Send first message */
	TRACE_DEBUG(FULL, "Handshake complete, now sending data over stream 0...");
	ret = tls_sctpw_send( tls_ctx, 0, MSG0, strlen(MSG0) );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "Sending data over encrypted link failed!");
		return -1;
	}
	
	/* To be continued... */
	sleep (2);
	
	/* Done, free the wrapper context */
	TRACE_DEBUG(FULL, "Test complete, now terminating the wrapper context...");
	ret = tls_sctpw_close( &tls_ctx );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "Failed to close the wrapper: %s", strerror(ret));
		return -1;
	}
	
	/* and close the socket */
	TRACE_DEBUG(FULL,"Closing the socket ...");
	ret = close(sock);
	if (ret != 0) {
		TRACE_DEBUG(FULL, "Failed to close the socket!");
		return -1;
	}
	
	return 0;
}

/* The server function */
int server_process()
{
	TRACE_ENTRY();
	
	int sock;
	int skcli;
	int streams;
	int ret = 0;
	tls_sctpw_ctx *tls_ctx = NULL;
	gnutls_dh_params_t dh_params;
	gnutls_anon_server_credentials_t anoncred;
	
	/* Initialize credentials */
	ret = gnutls_anon_allocate_server_credentials (&anoncred);
	if (ret != GNUTLS_E_SUCCESS) {
		TRACE_DEBUG(INFO,"Unable to initialize server credentials: %s", gnutls_strerror(ret));
		return -1;
	}
	ret = gnutls_dh_params_init (&dh_params);
	if (ret != GNUTLS_E_SUCCESS) {
		TRACE_DEBUG(INFO,"Unable to initialize Diffie-Hellman parameters: %s", gnutls_strerror(ret));
		return -1;
	}
	ret = gnutls_dh_params_generate2 (dh_params, 1024);
	if (ret != GNUTLS_E_SUCCESS) {
		TRACE_DEBUG(INFO,"Unable to generate Diffie-Hellman parameters: %s", gnutls_strerror(ret));
		return -1;
	}
	gnutls_anon_set_server_dh_params (anoncred, dh_params);
	
	/* Create the server socket */
	TRACE_DEBUG(FULL, "Creating server socket on port %d with %d streams ...", TEST_PORT, TEST_SRV_STR);
	ret = tls_sctp_create_server( &sock, TEST_PORT, TEST_SRV_STR );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "Failed to create server: %s", strerror(ret));
		return -1;
	}
	
	/* Accept a client connection. */
	TRACE_DEBUG(FULL, "Socket created, now waiting for a client connection ...");
	ret = tls_sctp_accept( sock, &skcli, &streams );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "Failed to accept connection: %s", strerror(ret));
		return -1;
	}
	
	/* Connection with a client is ready */
	TRACE_DEBUG(FULL, "Client connected, TLS can be run over %d streams", streams);
	
	/* test exchanging data without TLS */
	{
		TRACE_DEBUG(FULL, "Receiving data over SCTP streams 0 and 1 to test basic communication...");
		/* Client:
		ret = tls_sctp_send(sock, 0, MSG0, strlen(MSG0));
		ret = tls_sctp_send(sock, 1, MSG2, strlen(MSG2));
		*/
		tls_sctp_recvevents event;
		int strid;
		char * recvd;
		size_t len;
		int nmsg = 0;
		
		while (nmsg != 3) {
			ret = tls_sctp_recv(skcli, &event, &strid, (void **)&recvd, &len, 5);
			if (ret < 0) {
				TRACE_DEBUG(INFO,"tls_sctp_recv: %d", ret);
				return -1;
			}
			
			if (ret == 0) {
				TRACE_DEBUG(INFO,"tls_sctp_recv: event %d", event);
				continue;
			}
			
			if ((strid == 0) && (memcmp(recvd, MSG0, strlen(MSG0)) == 0))
				nmsg |= 1;
			if ((strid == 1) && (memcmp(recvd, MSG2, strlen(MSG2)) == 0))
				nmsg |= 2;
			
			free(recvd);
		}
	}	
	
	/* Now initialize the wrapper object */
	TRACE_DEBUG(FULL, "Data received; now creating wrapper context ...");
	ret = tls_sctpw_init( skcli, streams, &tls_ctx, 0, GNUTLS_CRD_ANON, (void **)&anoncred, set_dh_prio );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "Unable to create wrapper context: %s", strerror(ret));
		return -1;
	}
	
	/* TLS Handshake */
	TRACE_DEBUG(FULL, "Wrapper created, now initiating handshake ...");
	ret = tls_sctpw_handshake( tls_ctx, NULL, NULL );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "Handshake failed!");
		return -1;
	}
	
	/* Receive the first message sent by the client */
	{
		void * buf;
		int str;
		size_t len;
		
		TRACE_DEBUG(FULL, "Handshake complete, now receiving TLS-protected data...");
		ret = tls_sctpw_recv( tls_ctx, &str, &buf, &len );
		if (ret != 0) {
			TRACE_DEBUG(INFO, "Failed to receive data: %s", strerror(ret));
			return -1;
		}

		if ((str != 0) || (strcmp((char *)buf, MSG0))) {
			TRACE_DEBUG(INFO, "Received bad data (%d bytes over stream %d)...", len, str);
			return EBADMSG;
		}
	}
	
	/* Now send our first message */	
	TRACE_DEBUG(FULL, "Now sending data over stream 2 to client ...");
	ret = tls_sctpw_send( tls_ctx, 2, MSG1, strlen(MSG1) );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "Sending data over encrypted link failed!");
		return -1;
	}
	
	/* To be continued... */
	sleep (2);
	
	/* Finished, free the wrapper context */
	TRACE_DEBUG(FULL, "Test complete, now terminating the wrapper context...");
	ret = tls_sctpw_close( &tls_ctx );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "Failed to close the wrapper: %s", strerror(ret));
		return -1;
	}
	
	/* and close the socket */
	TRACE_DEBUG(FULL,"Closing the connected socket ...");
	ret = close(skcli);
	if (ret != 0) {
		TRACE_DEBUG(FULL, "Failed to close the socket!");
		return -1;
	}
	
	/* and close the server */
	TRACE_DEBUG(FULL,"Closing the server socket ...");
	ret = close(sock);
	if (ret != 0) {
		TRACE_DEBUG(FULL, "Failed to close the socket!");
		return -1;
	}
	
	return 0;
}

GCRY_THREAD_OPTION_PTHREAD_IMPL;

static void my_gnutls_log(int level, const char * str) 
{
	printf("%s[ gnutls * %d ] : %s", is_client?"                   [CLI] ":"[SRV] ", level, str);
	return;
}

/* Main */
int main(int argc, char * argv[])
{
	int ret = 0;
	pid_t client;
	
	TRACE_ENTRY("%d %p", argc, argv);
	
	/* Initialize gnutls */
 	gcry_control (GCRYCTL_SET_THREAD_CBS, &gcry_threads_pthread);
	gnutls_global_init ();
	gnutls_global_set_log_function(my_gnutls_log);
	gnutls_global_set_log_level(0);
	
	client = fork();
	if (client == 0) {
		/* Client process */
		is_client = 1;
		/* Wait for the server to start */
		sleep(2);
		/* Then start the client */
		return client_process();
	} else {
		if (client == (pid_t) -1) {
			TRACE_DEBUG(INFO, "Fork failed!");
			return -1;
		}
		
		/* Set a timeout just in case */
		alarm(20);
		
		/* server process */
		ret = server_process();

		TRACE_DEBUG(FULL, "Server process termination code: %d", ret);
		
		if (ret != 0) {
			/* Let's kill the client */
			kill(client, SIGTERM);
		}
		
		/* Wait for child to terminate */
		wait(&ret);
		
		TRACE_DEBUG(FULL, "Client process termination code: %x", ret);
	}
	
	TRACE_DEBUG(INFO, "%s is terminated.", argv[0]);
	return 0;
}

#endif /* STANDALONE_WRAPPER */


