/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* Peers facility.
 *
 * This file contains the code for the failover mechanism.
 *
 */
#include "waaad-internal.h"
#include "peer-internal.h"

static int _requeue_or_destroy(msg_t * msg)
{
	msg_data_t * hdr;
	
	TRACE_ENTRY("%p", msg);
	
	CHECK_FCT( msg_data(msg, &hdr) );
	
	/* Answers can only be sent to peers the request was received from, so answers are discarded */
	/* Also, messages that can not be routed are discarded */
	if (( hdr->msg_flags & CMD_FLAG_REQUEST ) && ( msg_is_routable( msg ) ) ) {
		/* Set the 'T' flag on the message */
		hdr->msg_flags |= CMD_FLAG_RETRANSMIT;
		
		/* Requeue the message in the global outgoing for routing daemon to pick it again */
		CHECK_FCT( meq_post(g_meq_outgoing, msg) );
	} else {
		/* There is no point in sending this message elsewhere, discard it */
		TRACE_DEBUG(FULL, "Message discarded during failover procedure (dump bellow)");
		msg_dump_walk(FULL, msg);
		CHECK_FCT( msg_free(msg, 1) );
	}
	return 0;
}
		
static int _requeue_or_requeue(msg_t * msg, uti_list_t * list)
{
	msg_data_t * hdr;
	
	TRACE_ENTRY("%p", msg);
	
	CHECK_FCT( msg_data(msg, &hdr) );
	
	/* Answers can only be sent to peers the request was received from, so answers are discarded */
	/* Also, messages that can not be routed are discarded */
	if (( hdr->msg_flags & CMD_FLAG_REQUEST ) && ( msg_is_routable( msg ) ) ) {
		/* Set the 'T' flag on the message */
		hdr->msg_flags |= CMD_FLAG_RETRANSMIT;
		
		/* Requeue the message in the global outgoing for routing daemon to pick it again */
		CHECK_FCT( meq_post(g_meq_outgoing, msg) );
	} else {
		/* We keep the message in the other list */
		uti_list_t * new = NULL;
		CHECK_MALLOC( new = (uti_list_t *) malloc(sizeof(uti_list_t)) );
		uti_list_init(new, msg);
		uti_list_insert_before(list, new);
	}
	return 0;
}
		
/* Requeue the messages that can be sent through another peer in the global queue */
int _peer_failover(_peer_t * peer, int freeloc)
{
	msg_t * msg = NULL;
	uti_list_t loc;
	
	uti_list_init(&loc, NULL);
	
	/* Deal with all requests in peer->p_out_q */
	while ( meq_tryget( peer->p_out_q, &msg ) == 0 ) {
		CHECK_FCT( _requeue_or_destroy( msg ) );
	}
	
	/* Deal with all requests in peer->p_sent */
	while ( !IS_LIST_EMPTY(&peer->p_sent) ) {
		uti_list_t * li = peer->p_sent.next; /* actually it's p_sent_t but it's OK like this */
		uti_list_unlink(li);
		if (freeloc) {
			CHECK_FCT( _requeue_or_destroy( (msg_t *)(li->o) ) );
		} else {
			CHECK_FCT( _requeue_or_requeue( (msg_t *)(li->o), &loc ) );
		}
		free(li);
	}
	
	/* Now requeue the local messages if needed */
	while ( !IS_LIST_EMPTY(&loc) ) {
		uti_list_t * li = loc.next;
		uti_list_unlink(li);
		uti_list_insert_before(&peer->p_sent, li);
	}
	
	/* The peer messages were cleaned properly */
	return 0;
}
